@@ -1,6 +1,6 @@
 /* Interface to Perl from OCaml.
  * Copyright (C) 2003 Merjis Ltd.
- * $Id: perl_c.c,v 1.12 2003/11/19 16:28:22 rich Exp $
+ * $Id: perl_c.c,v 1.19 2004/12/11 15:40:09 rich Exp $
  */
 
 #include <stdio.h>
@@ -9,61 +9,56 @@
 #include <unistd.h>
 #include <alloca.h>
 
+#define CAML_NAME_SPACE 1
+
 #include <caml/alloc.h>
 #include <caml/callback.h>
+#include <caml/custom.h>
 #include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/mlvalues.h>
 
-/* XXX This was required to avoid an error on my machine when loading the Perl
- * headers. Not clear why this is missing.
- */
-#define off64_t __off64_t
-
-/* XXX This is required by Perl >= 5.8.2. */
-#define __USE_GNU
-#include <crypt.h>
-
 #include <EXTERN.h>
 #include <perl.h>
 
 /* Perl requires the interpreter to be called literally 'my_perl'! */
 static PerlInterpreter *my_perl;
 
-/* Wrap up an arbitrary void pointer in an opaque OCaml object. */
-static value Val_voidptr (void *ptr);
-
 /* Get the concrete value from an optional field. */
 static value unoption (value option, value deflt);
 
+/* Wrap up an arbitrary void pointer in an opaque OCaml object. */
+static value Val_voidptr (void *ptr);
+
 /* Unwrap an arbitrary void pointer from an opaque OCaml object. */
 #define Voidptr_val(type,rv) ((type *) Field ((rv), 0))
 
+#if PERL4CAML_REFCOUNTING_EXPERIMENTAL
+
+/* Unwrap a custom block. */
+#define Xv_val(rv) (*((void **)Data_custom_val(rv)))
+
+/* Wrap up an SV, AV or HV in a custom OCaml object which will decrement
+ * the reference count on finalization.
+ */
+static value Val_xv (SV *sv);
+
+#else
+
+#define Xv_val(rv) Voidptr_val (SV, (rv))
+#define Val_xv(sv) Val_voidptr ((sv))
+
+#endif
+
 /* Hide Perl types in opaque OCaml objects. */
 #define Val_perl(pl) (Val_voidptr ((pl)))
 #define Perl_val(plv) (Voidptr_val (PerlInterpreter, (plv)))
-#define Val_sv(sv) (Val_voidptr ((sv)))
-#define Sv_val(svv) (Voidptr_val (SV, (svv)))
-#define Val_av(av) (Val_voidptr ((av)))
-#define Av_val(avv) (Voidptr_val (AV, (avv)))
-#define Val_hv(hv) (Val_voidptr ((hv)))
-#define Hv_val(hvv) (Voidptr_val (HV, (hvv)))
-
-CAMLprim value
-perl4caml_init (value unit)
-{
-  CAMLparam1 (unit);
-  PERL_SYS_INIT3 (NULL, NULL, NULL);
-  CAMLreturn (Val_unit);
-}
-
-CAMLprim value
-perl4caml_current_interpreter (value unit)
-{
-  CAMLparam1 (unit);
-  if (my_perl == 0) raise_not_found ();
-  CAMLreturn (Val_perl (my_perl));
-}
+#define Val_sv(sv) (Val_xv ((sv)))
+#define Sv_val(svv) ((SV *) Xv_val (svv))
+#define Val_av(av) (Val_xv ((SV *)(av)))
+#define Av_val(avv) ((AV *) Xv_val (avv))
+#define Val_hv(hv) (Val_xv ((SV *)(hv)))
+#define Hv_val(hvv) ((HV *) Xv_val (hvv))
 
 static void
 xs_init (pTHX)
@@ -75,61 +70,19 @@ xs_init (pTHX)
 }
 
 CAMLprim value
-perl4caml_create (value optargs, value unit)
+perl4caml_init (value unit)
 {
-  CAMLparam2 (optargs, unit);
-  CAMLlocal1 (args);
-  int argc, i;
-  char **argv;
-  static char *no_args[] = { "", "-w", "-e", "0" };
-
-  /* Arguments given? */
-  if (optargs == Val_int (0))	/* "None" */
-    {
-      argc = 4;
-      argv = no_args;
-    }
-  else				/* "Some args" where args is a string array. */
-    {
-      args = Field (optargs, 0);
-      argc = Wosize_val (args);
-      argv = alloca (argc * sizeof (char *));
-      for (i = 0; i < argc; ++i) argv[i] = String_val (Field (args, i));
-    }
+  CAMLparam1 (unit);
+  int argc = 4;
+  static char *argv[] = { "", "-w", "-e", "0", NULL };
 
+  PERL_SYS_INIT (&argc, &argv);
   my_perl = perl_alloc ();
   perl_construct (my_perl);
   PL_exit_flags |= PERL_EXIT_DESTRUCT_END;
-  perl_parse (my_perl, xs_init, argc, argv, NULL);
+  perl_parse (my_perl, xs_init, argc, argv, (char **) NULL);
   /*perl_run (my_perl);*/
 
-  CAMLreturn (Val_perl (my_perl));
-}
-
-CAMLprim value
-perl4caml_destroy (value plv)
-{
-  CAMLparam1 (plv);
-  PerlInterpreter *pl = Perl_val (plv);
-
-  perl_destruct (pl);
-  perl_free (pl);
-
-  /* Current interpreter? */
-  if (my_perl == pl) my_perl = 0;
-
-  CAMLreturn (Val_unit);
-}
-
-CAMLprim value
-perl4caml_set_context (value plv)
-{
-  CAMLparam1 (plv);
-  PerlInterpreter *pl = Perl_val (plv);
-
-  PERL_SET_CONTEXT (pl);
-  my_perl = pl;
-
   CAMLreturn (Val_unit);
 }
 
@@ -154,7 +107,7 @@ perl4caml_float_of_sv (value svv)
   CAMLparam1 (svv);
   SV *sv = Sv_val (svv);
   CAMLlocal1 (f);
-  f = copy_double (SvNV (sv));
+  f = caml_copy_double (SvNV (sv));
   CAMLreturn (f);
 }
 
@@ -175,7 +128,7 @@ perl4caml_string_of_sv (value svv)
   CAMLlocal1 (strv);
   str = SvPV (sv, len);
   /* XXX This won't work if the string contains NUL. */
-  strv = copy_string (str);
+  strv = caml_copy_string (str);
   CAMLreturn (strv);
 }
 
@@ -183,7 +136,7 @@ CAMLprim value
 perl4caml_sv_of_string (value strv)
 {
   CAMLparam1 (strv);
-  CAMLreturn (Val_sv (newSVpv (String_val (strv), string_length (strv))));
+  CAMLreturn (Val_sv (newSVpv (String_val (strv), caml_string_length (strv))));
 }
 
 CAMLprim value
@@ -198,7 +151,17 @@ CAMLprim value
 perl4caml_sv_undef (value unit)
 {
   CAMLparam1 (unit);
-  CAMLreturn (Val_sv (newSV (0)));
+  /*CAMLreturn (Val_sv (newSV (0)));*/
+  CAMLreturn (Val_sv (&PL_sv_undef));
+}
+
+CAMLprim value
+perl4caml_sv_is_undef (value svv)
+{
+  CAMLparam1 (svv);
+  SV *sv = Sv_val (svv);
+  CAMLreturn (!SvPOK (sv) && !SvIOK (sv) && SvTYPE (sv) == SVt_NULL
+	      ? Val_true : Val_false);
 }
 
 CAMLprim value
@@ -273,8 +236,8 @@ perl4caml_deref (value svv)
   CAMLlocal1 (rsvv);
   SV *sv = Sv_val (svv);
 
-  if (SvTYPE (sv) != SVt_RV)
-    invalid_argument ("deref: SV is not a reference");
+  if (!SvROK (sv))
+    caml_invalid_argument ("deref: SV is not a reference");
   switch (SvTYPE (SvRV (sv))) {
   case SVt_IV:
   case SVt_NV:
@@ -283,7 +246,7 @@ perl4caml_deref (value svv)
   case SVt_PVMG:
     break;
   default:
-    invalid_argument ("deref: SV is not a reference to a scalar");
+    caml_invalid_argument ("deref: SV is not a reference to a scalar");
   }
   rsvv = Val_sv (SvRV (sv));
   CAMLreturn (rsvv);
@@ -296,13 +259,13 @@ perl4caml_deref_array (value svv)
   CAMLlocal1 (ravv);
   SV *sv = Sv_val (svv);
 
-  if (SvTYPE (sv) != SVt_RV)
-    invalid_argument ("deref_array: SV is not a reference");
+  if (!SvROK (sv))
+    caml_invalid_argument ("deref_array: SV is not a reference");
   switch (SvTYPE (SvRV (sv))) {
   case SVt_PVAV:
     break;
   default:
-    invalid_argument ("deref_array: SV is not a reference to an array");
+    caml_invalid_argument ("deref_array: SV is not a reference to an array");
   }
   ravv = Val_av ((AV *) SvRV (sv));
   CAMLreturn (ravv);
@@ -315,13 +278,13 @@ perl4caml_deref_hash (value svv)
   CAMLlocal1 (rhvv);
   SV *sv = Sv_val (svv);
 
-  if (SvTYPE (sv) != SVt_RV)
-    invalid_argument ("deref_array: SV is not a reference");
+  if (!SvROK (sv))
+    caml_invalid_argument ("deref_hash: SV is not a reference");
   switch (SvTYPE (SvRV (sv))) {
   case SVt_PVHV:
     break;
   default:
-    invalid_argument ("deref_array: SV is not a reference to a hash");
+    caml_invalid_argument ("deref_hash: SV is not a reference to a hash");
   }
   rhvv = Val_hv ((HV *) SvRV (sv));
   CAMLreturn (rhvv);
@@ -434,7 +397,7 @@ perl4caml_av_get (value avv, value i)
   CAMLparam2 (avv, i);
   AV *av = Av_val (avv);
   SV **svp = av_fetch (av, Int_val (i), 0);
-  if (svp == 0) invalid_argument ("av_get: index out of bounds");
+  if (svp == 0) caml_invalid_argument ("av_get: index out of bounds");
   CAMLreturn (Val_sv (*svp));
 }
 
@@ -480,7 +443,7 @@ perl4caml_hv_set (value hvv, value key, value svv)
   HV *hv = Hv_val (hvv);
   SV *sv = Sv_val (svv);
   SvREFCNT_inc (sv);
-  if (hv_store (hv, String_val (key), string_length (key), sv, 0) == 0)
+  if (hv_store (hv, String_val (key), caml_string_length (key), sv, 0) == 0)
     SvREFCNT_dec (sv);
   CAMLreturn (Val_unit);
 }
@@ -490,8 +453,8 @@ perl4caml_hv_get (value hvv, value key)
 {
   CAMLparam2 (hvv, key);
   HV *hv = Hv_val (hvv);
-  SV **svp = hv_fetch (hv, String_val (key), string_length (key), 0);
-  if (svp == 0) raise_not_found ();
+  SV **svp = hv_fetch (hv, String_val (key), caml_string_length (key), 0);
+  if (svp == 0) caml_raise_not_found ();
   CAMLreturn (Val_sv (*svp));
 }
 
@@ -500,7 +463,7 @@ perl4caml_hv_exists (value hvv, value key)
 {
   CAMLparam2 (hvv, key);
   HV *hv = Hv_val (hvv);
-  bool r = hv_exists (hv, String_val (key), string_length (key));
+  bool r = hv_exists (hv, String_val (key), caml_string_length (key));
   CAMLreturn (r ? Val_true : Val_false);
 }
 
@@ -509,7 +472,7 @@ perl4caml_hv_delete (value hvv, value key)
 {
   CAMLparam2 (hvv, key);
   HV *hv = Hv_val (hvv);
-  hv_delete (hv, String_val (key), string_length (key), G_DISCARD);
+  hv_delete (hv, String_val (key), caml_string_length (key), G_DISCARD);
   CAMLreturn (Val_unit);
 }
 
@@ -540,7 +503,7 @@ perl4caml_get_sv (value optcreate, value name)
 
   create = unoption (optcreate, Val_false);
   sv = get_sv (String_val (name), create == Val_true ? TRUE : FALSE);
-  if (sv == NULL) raise_not_found ();
+  if (sv == NULL) caml_raise_not_found ();
 
   CAMLreturn (Val_sv (sv));
 }
@@ -554,7 +517,7 @@ perl4caml_get_av (value optcreate, value name)
 
   create = unoption (optcreate, Val_false);
   av = get_av (String_val (name), create == Val_true ? TRUE : FALSE);
-  if (av == NULL) raise_not_found ();
+  if (av == NULL) caml_raise_not_found ();
 
   CAMLreturn (Val_av (av));
 }
@@ -568,7 +531,7 @@ perl4caml_get_hv (value optcreate, value name)
 
   create = unoption (optcreate, Val_false);
   hv = get_hv (String_val (name), create == Val_true ? TRUE : FALSE);
-  if (hv == NULL) raise_not_found ();
+  if (hv == NULL) caml_raise_not_found ();
 
   CAMLreturn (Val_hv (hv));
 }
@@ -584,9 +547,9 @@ check_perl_failure ()
       STRLEN n_a;
       const char *err = SvPV (errsv, n_a);
 
-      errv = copy_string (err);
+      errv = caml_copy_string (err);
 
-      raise_with_arg (*caml_named_value ("perl4caml_perl_failure"), errv);
+      caml_raise_with_arg (*caml_named_value ("perl4caml_perl_failure"), errv);
     }
 }
 
@@ -701,10 +664,13 @@ perl4caml_call_array (value optsv, value optfnname, value arglist)
    */
   list = Val_int (0);
   for (i = 0; i < count; ++i) {
-    cons = alloc (2, 0);
+    SV *sv;
+
+    cons = caml_alloc (2, 0);
     Field (cons, 1) = list;
     list = cons;
-    Field (cons, 0) = Val_sv (newSVsv (POPs));
+    sv = newSVsv (POPs);
+    Field (cons, 0) = Val_sv (sv);
   }
 
   /* Restore the stack. */
@@ -875,10 +841,13 @@ perl4caml_call_method_array (value ref, value name, value arglist)
    */
   list = Val_int (0);
   for (i = 0; i < count; ++i) {
-    cons = alloc (2, 0);
+    SV *sv;
+
+    cons = caml_alloc (2, 0);
     Field (cons, 1) = list;
     list = cons;
-    Field (cons, 0) = Val_sv (newSVsv (POPs));
+    sv = newSVsv (POPs);
+    Field (cons, 0) = Val_sv (sv);
   }
 
   /* Restore the stack. */
@@ -1018,7 +987,7 @@ perl4caml_call_class_method_array (value classname, value name, value arglist)
    */
   list = Val_int (0);
   for (i = 0; i < count; ++i) {
-    cons = alloc (2, 0);
+    cons = caml_alloc (2, 0);
     Field (cons, 1) = list;
     list = cons;
     Field (cons, 0) = Val_sv (newSVsv (POPs));
@@ -1080,11 +1049,42 @@ perl4caml_call_class_method_void (value classname, value name, value arglist)
 static value
 Val_voidptr (void *ptr)
 {
-  value rv = alloc (1, Abstract_tag); /* XXX Is this correct? */
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+  rv = caml_alloc (1, Abstract_tag);
   Field(rv, 0) = (value) ptr;
-  return rv;
+  CAMLreturn (rv);
 }
 
+#if PERL4CAML_REFCOUNTING_EXPERIMENTAL
+
+static void
+xv_finalize (value v)
+{
+  SvREFCNT_dec ((SV *) Xv_val (v));
+}
+
+static struct custom_operations xv_custom_operations = {
+  "xv_custom_operations",
+  xv_finalize,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
+};
+
+static value
+Val_xv (SV *sv)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+  rv = caml_alloc_custom (&xv_custom_operations, sizeof (void *), 0, 1);
+  Xv_val (rv) = sv;
+  CAMLreturn (rv);
+}
+
+#endif /* PERL4CAML_REFCOUNTING_EXPERIMENTAL */
+
 static value
 unoption (value option, value deflt)
 {
